from rara_meta_extractor.tools.utils import load_txt, load_json
from rara_meta_extractor.constants.schemas import TEXT_CLASSIFIER_SCHEMA
from rara_meta_extractor.constants.data_classes import (
    TextBlock, IssueType,AuthorType, TitleType,  MetaYearField, MetaField, MergeMethod, TextPartLabel
)
from rara_tools.constants.meta_extractor import AUTHOR_ROLES_MAP
import importlib.resources
import logging


LOGGER = logging.getLogger("rara-meta-extractor")

METADATA_TEXT_BLOCKS = [TextBlock.METADATA, TextBlock.TITLE_PAGE]

# File paths
AUTHOR_FIELDS_FILE_PATH =  importlib.resources.files("rara_meta_extractor.data.fields") / "kata_author_fields.txt"
META_FIELDS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.fields") / "kata_meta_fields.txt"

META_EXTRACTOR_INSTRUCTIONS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.instructions") / "meta_extractor_instructions.txt"
TEXT_CLASSIFIER_INSTRUCTIONS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.instructions") / "text_classifier_instructions.txt"

META_SCHEMA_PATH = importlib.resources.files("rara_meta_extractor.data.schemas") / "meta_schema.json"

AUTHOR_ROLES_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.field_translations") / "author_roles.json"

EPUB_META_KEYS_FILE_PATH = importlib.resources.files("rara_meta_extractor.data.fields") / "epub_meta_keys.txt"

AUTHOR_ROLES_DICT = load_json(AUTHOR_ROLES_FILE_PATH)

EPUB_AUTHOR_ROLES_DICT = AUTHOR_ROLES_MAP

AUTHOR_FIELDS = load_txt(AUTHOR_FIELDS_FILE_PATH, to_list=True)
META_FIELDS = load_txt(META_FIELDS_FILE_PATH, to_list=True)

EPUB_META_KEYS = load_txt(EPUB_META_KEYS_FILE_PATH, to_list=True)


LLAMA_META_EXTRACTOR_INSTRUCTIONS = load_txt(META_EXTRACTOR_INSTRUCTIONS_FILE_PATH)
LLAMA_TEXT_CLASSIFIER_INSTRUCTIONS = load_txt(TEXT_CLASSIFIER_INSTRUCTIONS_FILE_PATH)

META_SCHEMA = load_json(META_SCHEMA_PATH)


LLAMA_TEXT_CLASSIFIER_URL = "http://dev-elastic1.texta.ee:8080"
LLAMA_META_EXTRACTOR_URL = "http://dev-elastic1.texta.ee:8080" # "http://localhost:8080"

FIELDS = META_FIELDS + AUTHOR_FIELDS


DEFAULT_INSTRUCTIONS = "This is a conversation between User and Llama - a metadata extrator with skills of professional cataloguer. Please extract the following fields: {0}"

KATA_INSTRUCTIONS = (
    f"This is a conversation between User and Llama - a metadata extrator with skills of professional cataloguer. " +
    f"Your only assignment is to extract relevant metadata and OUTPUT IT AS JSON DICT. " +
    f"THIS IS VERY IMPORTANT! Output it as JSON DICT or DIE. Fields you are trying to extract are the following: {FIELDS}\n"
)

META_EXTRACTOR_CONFIG = {
    "llama_host_url": LLAMA_META_EXTRACTOR_URL,
    "instructions": KATA_INSTRUCTIONS,
    "fields": [],
    "json_schema": {},
    "temperature": 0.1,
    "n_predict": 500
}

TEXT_CLASSIFIER_CONFIG = {
    "llama_host_url": LLAMA_TEXT_CLASSIFIER_URL,
    "instructions": LLAMA_TEXT_CLASSIFIER_INSTRUCTIONS,
    "fields": TEXT_CLASSIFIER_SCHEMA,
    "json_schema": {},
    "temperature": 0.0,
    "n_predict": 500
}

# For filtering false-positive authors extracted from EPUBs.
# Dont' allow names that contain
# more words than EPUB_AUTHOR_MAX_N_WORDS or
# more characters  than EPUB_AUTHOR_MAX_LENGTH
EPUB_AUTHOR_MAX_N_WORDS = 7
EPUB_AUTHOR_MAX_LENGTH = 50


# For filtering chapter names - the assumption is that
# if some required ratio (TABLE_OF_CONTENTS_MIN_UPPERCASE_RATIO)
# of chapters only contains uppercase characters and the average length
# of those chapter names is at least TABLE_OF_CONTENTS_MIN_UPPERCASE_AVG_LENGTH
# (to exclude names in only roman numerals etc), then chapter names in other casings
# are false positives
TABLE_OF_CONTENTS_MIN_UPPERCASE_RATIO = 0.3
TABLE_OF_CONTENTS_MIN_UPPERCASE_AVG_LENGTH = 10

MAX_LENGTH_PER_TEXT = 1700


FIELDS_TO_VALIDATE = [
    MetaField.TITLES, MetaField.AUTHORS, MetaField.PUBLISHER,
    MetaField.PUBLICATION_PLACE, MetaField.MANUFACTURER, MetaField.MANUFACTURE_PLACE
]

TITLE_TYPES_MAP = {
    TitleType.AUTHOR_WITHOUT_TITLE: 130,
    TitleType.NORMALIZED_TITLE: 240,
    TitleType.TITLE: 245,
    TitleType.PARALLEL_TITLE: 246,
    TitleType.ADDITIONAL_TITLE: 245,
    TitleType.METS_TITLE: 245
}


ISSUE_STYLE_MAP = {
    "newspaper": IssueType.NEWSPAPER,
    "ajalehed": IssueType.NEWSPAPER,
    "journal": IssueType.JOURNAL,
    "ajakirjad": IssueType.JOURNAL,
    "periodical": IssueType.PERIODICAL,
    "jätkväljanded": IssueType.PERIODICAL,
    "books": IssueType.BOOK,
    "raamatud": IssueType.BOOK
}
AUTHOR_TYPES_MAP = {"personal": AuthorType.PER}

DEFAULT_ISSUE_TYPE = IssueType.BOOK

TEXT_BLOCKS = [v for v in vars(TextBlock()).values()]
META_YEAR_FIELDS = [v for v in vars(MetaYearField()).values()]

META_FIELDS_WITH_MULTIPLE_VALUES = [MetaField.ISBN, MetaField.ISSN, MetaField.UDK, MetaField.UDC]

TITLE_KEYS = [
    MetaField.TITLE, MetaField.SUBTITLE, MetaField.TITLE_PART_NR,
    MetaField.TITLE_PART_NAME, MetaField.TITLE_VARFORM
]
META_KEYS_TO_IGNORE = TITLE_KEYS

METS_ALTO_DEFAULT_STYLE = ""

# Preferred extraction method. NB! This means that
# if one extractor does not detect this field, the output of other
# is still used.
DEFAULT_EPUB_MERGE_METHOD = MergeMethod.EPUB_SOFT
DEFAULT_LLAMA_MERGE_METHOD = MergeMethod.LLAMA_HARD


EPUB_FIELD_METHOD_MAP = {
    MetaField.TITLES: MergeMethod.COMBO,
    MetaField.AUTHORS: MergeMethod.COMBO,
    MetaField.PUBLISHER: MergeMethod.LLAMA_SOFT,
    MetaField.ISBN: MergeMethod.REGEX,
    MetaField.PUBLICATION_DATE: MergeMethod.EPUB_SOFT
}

LLAMA_FIELD_METHOD_MAP = {
    MetaField.ISBN: MergeMethod.REGEX
}



from typing import List, Tuple

from rara_meta_extractor.config import LOGGER
from rara_meta_extractor.constants.data_classes import TextPartLabel, MetaField
from rara_meta_extractor.text_part_classifiers.base_text_part_classifier import BaseTextPartClassifier


class MetsAltoTextPartClassifier(BaseTextPartClassifier):
    def __init__(self, extra_logging: dict = None):
        super().__init__()
        self.section_types: dict = {
            TextPartLabel.TABLE_OF_CONTENTS: "TABLE_OF_CONTENTS",
            TextPartLabel.CONCLUSION: "CONCLUSION",
            TextPartLabel.ABSTRACT: "ABSTRACT"
        }
        self.extra_logging = extra_logging or {}
        self.used_sections: list = []
        self.conclusion_upper_keywords: list = []

    def _metsalto_parsed_sections_to_toc(self, metsalto_parsed_texts: list) -> Tuple[list, list]:
        """ Putting together table of contents using metsalto
        section titles and authors from section meta.
        """
        toc = []

        if metsalto_parsed_texts:
            for item in metsalto_parsed_texts:
                sec_data = {}
                if MetaField.TITLES in item:
                    sec_data["title"] = item[MetaField.TITLES]
                if MetaField.AUTHORS in item:
                    sec_data["names"] = [
                        name["name"]
                        for name in item[MetaField.AUTHORS]
                    ]
                toc.append(sec_data)
        return toc

    def _get_section_types(self,
                           digitized_texts: List[dict], search_type: str
                           ) -> List[dict]:
        """ Getting text by using section types.
        """
        found_items = []
        for section in digitized_texts:
            if "section_type" in section:
                if section["section_type"] == self.section_types[search_type]:
                    if search_type == TextPartLabel.TABLE_OF_CONTENTS:
                        found_items.append(self._remove_double_toc(section["text"]))
                        self.used_sections.append(section["sequence_nr"])
                    elif search_type == TextPartLabel.CONCLUSION:
                        found_items.append(
                            {
                                "sequence_nr": section["sequence_nr"],
                                "text_value": section["text"],
                                "task_language": section["language"]
                            }
                        )
                        self.used_sections.append(section["sequence_nr"])
                    else:
                        found_items.append(
                            {
                                "text_value": section["text"],
                                "task_language": section["language"]
                            }
                        )
                        self.used_sections.append(section["sequence_nr"])
        if search_type == TextPartLabel.CONCLUSION:
            found_items = self._postprocess_concl(found_items)
        return found_items

    def get_conclusions(self, digitized_texts: List[dict]) -> List[dict]:
        """ Using different methods to find conclusions.
        """
        LOGGER.debug(f"Searching for conclusions.", extra=self.extra_logging)
        concl = self._get_section_types(
            digitized_texts=digitized_texts,
            search_type=TextPartLabel.CONCLUSION
        )
        if not concl:
            concl = self._look_through_section_titles(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.CONCLUSION
            )
        if not concl:
            concl = self._look_through_texts(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.CONCLUSION
            )
        return concl

    def get_tables_of_content(self, metsalto_sections, digitized_texts: List[dict]) -> List[dict]:
        """ Using diferents methods to find tables of content.
        """
        LOGGER.debug(f"Searching for tables of content.", extra=self.extra_logging)
        # Using different methods to find tables of content for mets-alto files
        # since mets_alto usually has authors names associated with chapters,
        # we first try to fetch section titles and author names by themselves
        toc = self._metsalto_parsed_sections_to_toc(metsalto_sections)
        if not toc:
            # then we try to extract the text through section types
            toc = self._get_section_types(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.TABLE_OF_CONTENTS
            )
        if not toc:
            # after this we look at where the toc should be according
            # to section titles and extract the text
            toc = self._look_through_section_titles(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.TABLE_OF_CONTENTS
            )
        if not toc:
            # after this we look at the text itself using keywords
            toc = self._look_through_texts(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.TABLE_OF_CONTENTS
            )
        return toc

    def get_abstracts(self, digitized_texts: List[dict]) -> List[dict]:
        """ Using different methods to find abstracts.
        """
        LOGGER.debug(f"Searching for abstracts.", extra=self.extra_logging)
        abstracts = self._get_section_types(
            digitized_texts=digitized_texts,
            search_type=TextPartLabel.ABSTRACT
        )
        if not abstracts:
            abstracts = self._look_through_section_titles(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.ABSTRACT
            )
        if not abstracts:
            abstracts = self._look_through_texts(
                digitized_texts=digitized_texts,
                search_type=TextPartLabel.ABSTRACT
            )
        return abstracts

    def get_parts_of_text(self, digitized_texts: List[dict], mets_alto_meta: dict) -> dict:
        """ Getting results for mets-alto files

        Parameters
        -----------
        digitized_texts: List[dict]
            `texts` from Digitizer output.
        epub_meta: dict
            Output of `MetsAltoMetaExtractor.extract_meta()`.

        Returns
        -----------
        dict
            TODO
        """
        LOGGER.info(f"Classifying METS/ALTO text sections.", extra=self.extra_logging)
        # TODO: move "sections" key into constants! but same
        # with this key in mets alto meta extractor!
        metsalto_sections = mets_alto_meta.get(MetaField.SECTIONS, [])

        tables_of_content = self.get_tables_of_content(
            metsalto_sections=metsalto_sections,
            digitized_texts=digitized_texts
        )
        conclusions = self.get_conclusions(
            digitized_texts=digitized_texts
        )
        abstracts = self.get_abstracts(
            digitized_texts=digitized_texts
        )

        results = self._results_correct_fields(
            {
                "table_of_contents": tables_of_content,
                "conclusions": conclusions,
                "abstracts": abstracts
            }
        )
        return results



from rara_tools.parsers.mets_alto_parsers.mets_alto_parser import DocumentMeta
from rara_meta_extractor.constants.data_classes import  TitleType, AuthorNameOrder
from rara_meta_extractor.tools.meta_formatter import Author, Title
from copy import deepcopy
from typing import List


class MetsAltoOutputEnricher:
    """ Enriches parsed METS/ALTO with necessary
    title and author fields.
    """
    def __init__(self, doc: dict):
        self.original_doc: dict = doc

    @staticmethod
    def enrich_titles(titles: List[dict], simple: bool = False) -> List[dict]:
        """ Add information expected by Kratt CORE to each title.
        """
        enriched_titles =  []

        # The current Title formatter is a bit illogical
        # as it takes in a list of titles instead of a single one.
        # However, we would like to add language separately for each title,
        # so we just call the formatter multiple times
        for title in titles:
            new_titles = Title(
                titles = [title.get("title")],
                lang = title.get("language"),
                part_number = title.get("part_number"),
                title_type=TitleType.METS_TITLE
            ).to_dicts(simple=simple)
            enriched_titles.extend(new_titles)
        return enriched_titles

    @staticmethod
    def enrich_authors(authors: List[dict]) -> List[dict]:
        """ Add information expected by Kratt CORE to each author.
        """
        enriched_authors = []

        for i, author in enumerate(authors):
            name = f"{author.get('first_name')} {author.get('last_name')}"
            author_name_order = AuthorNameOrder.FIRST_NAME_FIRST
            #is_primary = True if i == 0 else False

            new_author = Author(
                name=name,
                role=author.get("role"),
                author_type=author.get("type"),
                author_name_order=author_name_order,
                is_primary=True,
                map_role=False
            ).to_dict()
            enriched_authors.append(new_author)
        return enriched_authors

    def enrich(self, simple: bool = False) -> dict:
        """ Add information expected by Kratt CORE to titles & authors
        (the whole document + sections).
        """
        extracted_meta: dict = deepcopy(self.original_doc)

        # Main titles & authors
        titles = extracted_meta.pop("title", [])
        authors = extracted_meta.pop("authors", [])

        enriched_titles = MetsAltoOutputEnricher.enrich_titles(titles=[titles], simple=simple)
        enriched_authors = MetsAltoOutputEnricher.enrich_authors(authors=authors)
        if titles:
            extracted_meta["titles"] = enriched_titles
        if authors:
            extracted_meta["authors"] = enriched_authors

        # Section titles & authors
        sections = extracted_meta.pop("sections", [])
        enriched_sections = []
        for section in sections:
            titles = section.pop("titles", [])
            merged_titles = section.pop("merged_titles", [])
            merged_titles = [title for title in merged_titles if title.get("title")]
            authors = section.pop("authors", [])

            authors = [author for author in authors if author.get("first_name") or author.get("")]

            if not authors and not merged_titles:
                continue

            enriched_authors = MetsAltoOutputEnricher.enrich_authors(authors=authors)
            enriched_titles = MetsAltoOutputEnricher.enrich_titles(titles=merged_titles, simple=simple)

            section["titles"] = enriched_titles
            section["authors"] = enriched_authors

            enriched_sections.append(section)
        extracted_meta["sections"] = enriched_sections
        return extracted_meta


class MetsAltoMetaExtractor:
    def __init__(self, mets_alto_metadata: List[dict], language: str, texts: List[dict]):
        self.digitizer_output: dict = self.reconstruct_digitizer_output(
            mets_alto_metadata=mets_alto_metadata,
            language=language,
            texts=texts
        )
        self.document_meta: dict = DocumentMeta(self.digitizer_output).to_dict()
        self.output_enricher = MetsAltoOutputEnricher(self.document_meta)

    def reconstruct_digitizer_output(
            self, mets_alto_metadata: List[str] = [], language: str = "", texts: List[dict] = []
        ) -> dict:
        """ Partially reconstructs digitizer output as the new Mets/Alto meta extractor
        expects this format.
        """
        digitizer_output = {
            "doc_meta": {
                "mets_alto_metadata": mets_alto_metadata,
                "languages": [{"language": language}]
            },
            "texts": texts
        }
        return digitizer_output

    def remove_empty(self, meta: dict) -> dict:
        """ Removes empty meta fields.
        """
        sections = meta.pop("sections", [])
        cleaned_sections = []
        for section in sections:
            new_section = {}
            for key, value in section.items():
                if value:
                    new_section[key] = value
            cleaned_sections.append(new_section)

        new_meta = {}
        for key, value in meta.items():
            if value:
                new_meta[key] = value
        new_meta["sections"] = cleaned_sections
        return new_meta

    def reformat_fields(self, meta: dict) -> dict:
        """ Rename fields & reformat data structures.
        """
        isbn = meta.pop("isbn", "")
        issn = meta.pop("issn", "")
        content_type = meta.pop("resource_type", "")
        if isbn:
            meta["isbn"] = [isbn]
        if issn:
            meta["issn"] = [issn]
        if content_type:
            meta["content_type"] = content_type
        return meta

    def extract_meta(self, simple: bool = False) -> dict:
        """ Extract metadata. Use mainly DocumentMeta class,
        but reformat / enrich content of some fields like titles and authors.
        """
        extracted_meta = self.output_enricher.enrich(simple=simple)
        extracted_meta = self.remove_empty(extracted_meta)
        extracted_meta = self.reformat_fields(extracted_meta)
        return extracted_meta



from collections import defaultdict
from typing import List, NoReturn

from rara_meta_extractor.config import (
    LOGGER
)
from rara_meta_extractor.constants.data_classes import (
    MetaField, ExtractorType, MergeMethod
)
from rara_meta_extractor.regex_extractors import ISBNRegexExtractor
from rara_meta_extractor.tools.meta_formatter import MetaValidator


class EpubLlamaMerger:
    """ For merging meta detected with LlamaExtractor and/or EPUBMetaExtractor.
    """

    def __init__(
            self,
            epub_meta: dict,
            llama_meta: dict,
            text: str,
            field_method_map: dict,
            default_merge_method: str,
            extra_logging: dict = None,
    ) -> NoReturn:
        self.epub_meta: dict = epub_meta
        self.llama_meta: dict = llama_meta
        self.text: str = text
        self.llama_epub_field_map: dict = field_method_map
        self.default_merge_method: str = default_merge_method
        self.extra_logging: dict = extra_logging or {}
        self.meta_validator = MetaValidator()
        self.isbn_regex_extractor = ISBNRegexExtractor()

    def _is_valid_key(self, field: str) -> bool:
        return self.meta_validator._is_valid_key(field)

    @property
    def epub_fields(self) -> List[str]:
        return list(self.epub_meta.keys())

    @property
    def llama_fields(self) -> List[str]:
        return list(self.llama_meta.keys())

    @property
    def common_fields(self) -> List[str]:
        return list(set(self.epub_fields).intersection(set(self.llama_fields)))

    @property
    def union_fields(self) -> List[str]:
        fields = list(set(self.epub_fields).union(set(self.llama_fields)))
        fields = [field for field in fields if self._is_valid_key(field)]
        return fields

    @property
    def epub_unique_fields(self) -> List[str]:
        return list(set(self.epub_fields) - set(self.llama_fields))

    @property
    def llama_unique_fields(self) -> List[str]:
        return list(set(self.llama_fields) - set(self.epub_fields))

    @property
    def epub_titles(self) -> List[dict]:
        return self.epub_meta.get(MetaField.TITLES, [])

    @property
    def llama_titles(self) -> List[dict]:
        return self.llama_meta.get(MetaField.TITLES, [])

    @property
    def epub_authors(self) -> List[dict]:
        return self.epub_meta.get(MetaField.AUTHORS, [])

    @property
    def llama_authors(self) -> List[dict]:
        return self.llama_meta.get(MetaField.AUTHORS, [])

    @property
    def epub_authors_str(self) -> List[dict]:
        return [author.get("name") for author in self.epub_authors]

    @property
    def llama_authors_str(self) -> List[dict]:
        return [author.get("name") for author in self.llama_authors]

    @property
    def unique_llama_authors(self) -> List[dict]:
        unique = set(self.llama_authors_str) - set(self.epub_authors_str)
        filtered = [author for author in self.llama_authors if author.get("name") in unique]
        return filtered

    @property
    def unique_epub_authors(self) -> List[dict]:
        unique = set(self.epub_authors_str) - set(self.llama_authors_str)
        filtered = [author for author in self.epub_authors if author.get("name") in unique]
        return filtered

    @property
    def common_authors(self) -> List[str]:
        common = list(set(self.epub_authors_str).intersection(set(self.llama_authors_str)))
        return common

    def _get_titles_map(self, meta: dict) -> dict:
        titles = meta.get(MetaField.TITLES, [])
        titles_map = defaultdict(list)

        for title in titles:
            key = title.get("title_type_int")
            titles_map[key].append(title)

        return titles_map

    def _is_capitalized(self, title: str) -> bool:
        title_words = [word.strip() for word in title.split() if word.strip()]
        for word in title_words:
            if len(word) > 3 and word[0] != word[0].upper():
                return False
        return True

    def _select_title(self, key: str, llama_title: dict, epub_title: dict) -> dict:
        llama_title_str = llama_title.get("title", "")
        epub_title_str = epub_title.get("title", "")

        if llama_title_str and not epub_title_str:
            title = llama_title
        elif epub_title_str and not llama_title_str:
            title = epub_title
        elif llama_title_str.lower() == epub_title_str.lower():
            if self._is_capitalized(epub_title_str) or not self._is_capitalized(llama_title_str):
                title = epub_title
            else:
                title = llama_title
        elif llama_title_str.lower() in epub_title_str.lower():
            if ":" in epub_title_str and ":" not in llama_title_str:
                title = epub_title
            else:
                title = llama_title
        else:
            title = epub_title

        if title.get("title") == epub_title_str:
            LOGGER.debug(
                f"Title key = {key}. Using title extracted with EPUB-Extractor ({epub_title_str}). " \
                f"Ignoring title extracted with Llama-Extractor ({llama_title_str}).",
                extra=self.extra_logging
            )
        else:
            LOGGER.debug(
                f"Title key = {key}. Using title extracted with Llama-Extractor ({llama_title_str}). " \
                f"Ignoring title extracted with EPUB-Extractor ({epub_title_str}).",
                extra=self.extra_logging
            )
        return title

    def _get_unique_titles(self, titles_map: dict, common_keys: List[str], extractor_type: str) -> List[dict]:
        unique_titles = []
        for key, titles in titles_map.items():
            if key not in common_keys:
                LOGGER.debug(
                    f"Adding titles unique to extraction method '{extractor_type}'into results: {titles}",
                    extra=self.extra_logging
                )
                unique_titles.extend(titles)
        return unique_titles

    def _merge_titles(self):
        LOGGER.info(f"Comparing titles extracted with LLama-Extractor and EPUB-Extractor...", extra=self.extra_logging)
        llama_titles_map = self._get_titles_map(self.llama_meta)
        epub_titles_map = self._get_titles_map(self.epub_meta)

        common_keys = list(set(llama_titles_map.keys()).intersection(epub_titles_map.keys()))
        titles = []
        epub_title_added = False

        for key in common_keys:
            # Take just the first title  with that type,
            # not sure, if it is even possible to have more
            # than one title per title type
            epub_title = epub_titles_map.get(key)[0]
            llama_title = llama_titles_map.get(key)[0]
            final_title = self._select_title(key=key, llama_title=llama_title, epub_title=epub_title)
            titles.append(final_title)
            epub_title_added = True

        llama_unique_titles = self._get_unique_titles(llama_titles_map, common_keys, ExtractorType.LLAMA)
        epub_unique_titles = self._get_unique_titles(epub_titles_map, common_keys, ExtractorType.EPUB)

        if epub_unique_titles:
            epub_title_added = True
        titles.extend(epub_unique_titles)

        # Add titles unique to Llama ONLY IF no EPUB titles have been added
        # in order to avoid weird combinations
        if not epub_title_added:
            titles.extend(llama_unique_titles)
        return titles

    def _merge_authors(self):
        merged_authors = []
        merged_authors.extend(self.unique_llama_authors)
        merged_authors.extend(self.unique_epub_authors)

        for author in self.common_authors:
            roles = set()
            for llama_author in self.llama_authors:
                if llama_author.get("name") == author and llama_author.get("role") not in roles:
                    roles.add(llama_author.get("role"))
                    merged_authors.append(llama_author)

            for epub_author in self.epub_authors:
                if epub_author.get("name") == author and epub_author.get("role") not in roles:
                    roles.add(epub_author.get("role"))
                    merged_authors.append(epub_author)
        return merged_authors

    def _run_value_selection(self, field: str, merge_method: str) -> None | List[dict] | List[str] | str:
        value = None
        LOGGER.info(
            f"Running value selection for field '{field.upper()}' with merge method = {merge_method}.",
            extra=self.extra_logging
        )
        if merge_method in [MergeMethod.EPUB_SOFT, MergeMethod.EPUB_HARD]:
            value = self.epub_meta.get(field)
        elif merge_method in [MergeMethod.LLAMA_SOFT, MergeMethod.LLAMA_HARD]:
            value = self.llama_meta.get(field)
        if merge_method == MergeMethod.EPUB_SOFT and not value:
            LOGGER.debug(
                f"Could not detect the value for field '{field.upper()}' from EPUB meta, " \
                f"trying to retrieve it from Llama meta.",
                extra=self.extra_logging
            )
            value = self.llama_meta.get(field)
        if merge_method == MergeMethod.LLAMA_SOFT and not value:
            LOGGER.debug(
                f"Could not detect the value for field '{field.upper()}' from Llama meta, " \
                f"trying to retrieve it from EPUB meta.",
                extra=self.extra_logging
            )
            value = self.epub_meta.get(field)
        return value

    @property
    def authors(self) -> List[dict]:
        LOGGER.debug(f"EPUB authors: {self.epub_authors}", extra=self.extra_logging)
        LOGGER.debug(f"Llama authors: {self.llama_authors}", extra=self.extra_logging)
        merge_method = self.llama_epub_field_map.get(MetaField.AUTHORS, MergeMethod.COMBO)
        LOGGER.info(f"Running AUTHORS selection with merge method = {merge_method}.", extra=self.extra_logging)
        if merge_method == MergeMethod.COMBO:
            authors = self._merge_authors()

        else:
            authors = self._run_value_selection(field=MetaField.AUTHORS, merge_method=merge_method)

        roles = [author.get("role").lower() for author in authors]

        if "autor" in roles:
            authors = [author for author in authors if author.get("role").lower() != "unknown"]
        # If we could not detect an author, assume that the one with unknown role might be one
        elif "unknown" in roles:
            for author in authors:
                if author.get("role") == "unknown":
                    author["role"] = "Autor"
        return authors

    @property
    def titles(self) -> List[dict]:
        merge_method = self.llama_epub_field_map.get(MetaField.TITLES, MergeMethod.COMBO)
        LOGGER.info(f"Running TITLES selection with merge method = {merge_method}.", extra=self.extra_logging)
        if merge_method == MergeMethod.COMBO:
            titles = self._merge_titles()
        else:
            titles = self._run_value_selection(field=MetaField.TITLES, merge_method=merge_method)
        return titles

    @property
    def isbn(self) -> List[str]:
        merge_method = self.llama_epub_field_map.get(MetaField.ISBN, MergeMethod.REGEX)
        LOGGER.info(f"Running ISBN selection with merge method = {merge_method}.", extra=self.extra_logging)
        if merge_method == MergeMethod.REGEX:

            isbn = self.isbn_regex_extractor.extract(self.text)
            LOGGER.debug(f"ISBNs detected with REGEX: {isbn}", extra=self.extra_logging)
        else:
            isbn = self._run_value_selection(field=MetaField.ISBN, merge_method=merge_method)
        return isbn

    def _get_merge_method(self, field: str) -> str:
        field_1 = "_".join(field.split())
        field_2 = " ".join(field.split("_"))
        merge_method = self.llama_epub_field_map.get(field_1, None)
        if not merge_method:
            merge_method = self.llama_epub_field_map.get(field_2, self.default_merge_method)
        return merge_method

    def merge(self) -> dict:
        LOGGER.info(f"EPUB meta to merge: {self.epub_meta}", extra=self.extra_logging)
        LOGGER.info(f"Llama meta to merge: {self.llama_meta}", extra=self.extra_logging)
        merged_meta = {}

        for field in self.union_fields:
            if field == MetaField.TITLES:
                merged_meta[field] = self.titles
            elif field == MetaField.AUTHORS:
                merged_meta[field] = self.authors
            elif field == MetaField.ISBN:
                merged_meta[field] = self.isbn
            else:
                merge_method = self._get_merge_method(field)
                LOGGER.info(
                    f"Running {field.upper()} selection with merge method = {merge_method}.",
                    extra=self.extra_logging
                )
                merged_meta[field] = self._run_value_selection(field=field, merge_method=merge_method)

        return merged_meta



from rara_meta_extractor.tools.meta_formatter import Title
from typing import List

import pytest
import os

TEST_SKIP_TITLES = [
    ("Andromeda", 0),
    ("...And Then There Were None", 3),
    ("+ Teised", 2),
    ("Ja nii edasi...", 0),
    ("Miks me elame?", 0),
    ("A Fun Story", 2),
    ("The Great Escape", 4),
    ("... The People", 8)
]

def test_skip_calculation():
    for test_title in TEST_SKIP_TITLES:
        title, expected_skip = test_title
        true_skip = Title.find_skip(title)
        assert expected_skip == true_skip



from rara_meta_extractor.mets_alto_meta_extractor import MetsAltoMetaExtractor
from rara_meta_extractor.tools.utils import jl_generator
from typing import List
from pprint import pprint

import pytest
import os

def load_testset(file_path: str) -> List[dict]:
    testset = []
    for doc in jl_generator(file_path):
        doc_id = doc.get("dir_name")
        data = doc.get("digitized_data")
        doc_meta = data.get("doc_meta")
        languages = doc_meta.get("languages")
        language = languages[0].get("language", "") if languages else ""
        mets_alto_meta = doc_meta.get("mets_alto_metadata")
        digitized_texts = data.get("texts")
        testset.append(
            {
                "mets_alto_metadata": mets_alto_meta,
                "texts": digitized_texts,
                "language": language
            }
        )
    return testset


ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_FILE = os.path.join(ROOT_DIR, "tests", "test_data", "mets_alto_meta", "digitized_mets_alto.jl")
TESTSET = load_testset(TEST_FILE)

def test_mets_alto_extractor():
    for doc_batch in TESTSET:
        mets_alto_extractor = MetsAltoMetaExtractor(**doc_batch)
        meta = mets_alto_extractor.extract_meta(simple=False)
        pprint(meta)
        assert meta




from copy import deepcopy
from typing import List

import regex as re

from rara_meta_extractor.config import (
    LOGGER, FIELDS_TO_VALIDATE
)
from rara_meta_extractor.constants.data_classes import MetaField


class LlamaValidator:
    def __init__(self, fields_to_validate: List[str] = FIELDS_TO_VALIDATE, extra_logging: dict = None):
        self.fields_to_validate: List[str] = self._format_fields(fields_to_validate)
        self.extra_logging: dict = extra_logging or {}

    def _format_fields(self, fields: List[str]) -> List[str]:
        """ Make sure that all field names contain "_" instead of " " as word separators.
        """
        new_fields = []
        for field in fields:
            tokens = field.split()
            new_field = "_".join(tokens)
            new_fields.append(new_field)
        return new_fields

    def filter_false_positives(self, text: str, llama_output: dict):
        LOGGER.info(f"Filtering Llama output: {llama_output}", extra=self.extra_logging)
        filtered_output = deepcopy(llama_output)
        text = text.lower()
        for field in self.fields_to_validate:
            LOGGER.debug(f"Validating content for field '{field}'...", extra=self.extra_logging)
            if field == MetaField.AUTHORS:
                key = "name"
            elif field == MetaField.TITLES:
                key = "title"

            values = llama_output.get(field, [])
            if not values:
                LOGGER.debug(
                    f"Values for field '{field}' are not detected with Llama-Extractor or have already been filtered out.",
                    extra=self.extra_logging
                )
            if isinstance(values, str):
                val_value = values
                tokens = [re.escape(t.strip()) for t in val_value.lower().split() if t.strip()]
                search_value = r"\s*".join(tokens)
                if re.search(search_value, text):
                    filtered_output[field] = val_value
                    LOGGER.debug(
                        f"Detected value '{val_value}' (field={field}) from the original text. Keeping it!",
                        extra=self.extra_logging
                    )

                else:
                    filtered_output.pop(field, "")
                    LOGGER.debug(
                        f"Could not detect {field.upper()} value '{val_value}' from the original text. " \
                        f"Removing it from the output.",
                        extra=self.extra_logging
                    )

            elif isinstance(values, list):
                filtered_list = []
                for value in values:
                    val_value = ""
                    if isinstance(value, dict):
                        val_value = value.get(key)
                    elif isinstance(value, str):
                        val_value = value
                    if val_value:
                        tokens = [re.escape(t.strip()) for t in val_value.lower().split() if t.strip()]
                        search_value = r"\s*".join(tokens)
                        if re.search(search_value, text):
                            filtered_list.append(value)
                            LOGGER.debug(
                                f"Detected value '{val_value}' (field={field}) from the original text. Keeping it!",
                                extra=self.extra_logging
                            )
                        else:
                            LOGGER.debug(
                                f"Could not detect {field.upper()} value '{val_value}' from the original text. " \
                                f"Removing it from the output.",
                                extra=self.extra_logging
                            )
                if filtered_list:
                    filtered_output[field] = filtered_list
                else:
                    filtered_output.pop(field, "")
            else:
                LOGGER.debug(f"Values have unexpected type: {type(values)}.", extra=self.extra_logging)

        LOGGER.info(f"Final filtered Llama-Extractor output: {filtered_output}", extra=self.extra_logging)
        return filtered_output



from rara_meta_extractor.constants.data_classes import MetaField, TextBlock

TEXT_BLOCKS = [v for v in vars(TextBlock()).values()]

TEXT_CLASSIFIER_SCHEMA = [
    {
        "name": MetaField.TEXT_TYPE,
        "item_restrictions": {
            "enum": TEXT_BLOCKS,
        },
        "list_restrictions": {
            "maxItems": 1,
            "minItems": 1
        }
    }
]

META_SCHEMA = [
    {
        "name": "title",
        "item_restrictions": {}
    },
    {
        "name": "original title",
        "item_restrictions": {}
    },
    {
        "name": "title part number",
        "item_restrictions": {}
    },
    {
        "name": "edition",
        "item_restrictions": {}
    },
    {
        "name": "publisher",
        "item_restrictions": {}
    },
    {
        "name": "publication year",
        "item_restrictions": {
            "maxLength": 4,
            "minLength": 4,
            "type": "integer"
        }
    },
    {
        "name": "publication place",
        "item_restrictions": {}
    },
    {
        "name": "manufacture name",
        "item_restrictions": {}
    },
    {
        "name": "manufacture place",
        "item_restrictions": {}
    },
    {
        "name": "manufacture year",
        "item_restrictions": {
            "maxLength": 4,
            "minLength": 4,
            "type": "integer"
        }
    },
    {
        "name": "distribution name",
        "item_restrictions": {}
    },
    {
        "name": "distribution place",
        "item_restrictions": {}
    },
    {
        "name": "distribution year",
        "item_restrictions": {
            "maxLength": 4,
            "minLength": 4,
            "type": "integer"
        }
    },
    {
        "name": "publication place",
        "item_restrictions": {}
    },
    {
        "name": "copyright year",
        "item_restrictions": {
            "maxLength": 4,
            "minLength": 4,
            "type": "integer"
        }
    },
    {
        "name": "country",
        "item_restrictions": {}
    },
    {
        "name": "isbn",
        "item_restrictions": {
            "maxLength": 13,
            "minLength": 13,
            "type": "integer"
        }
    },
    {
        "name": "issn",
        "item_restrictions": {
            "maxLength": 8,
            "minLength": 8,
            "type": "integer"
        }
    }
]

AUTHORS_SCHEMA = [
    {
        "name": "author",
        "item_restrictions": {}
    },
    {
        "name": "story adaptor",
        "item_restrictions": {}
    },
    {
        "name": "foreword author",
        "item_restrictions": {}
    },
    {
        "name": "translator",
        "item_restrictions": {}
    },
    {
        "name": "illustrator",
        "item_restrictions": {}
    },
    {
        "name": "editor",
        "item_restrictions": {}
    },
    {
        "name": "designer",
        "item_restrictions": {}
    },
    {
        "name": "photographer",
        "item_restrictions": {}
    },
    {
        "name": "language editor",
        "item_restrictions": {}
    }
]



import requests
import json
from typing import List, Union

#https://github.com/ggml-org/llama.cpp/tree/master/grammars
class JSONSchemaGenerator:
    def __init__(self):
        self.keywords_key = "keywords"

    def generate_keyword_schema(
        self, keywords: List[str] = [],
        min_keywords: int | None = None,
        max_keywords: int | None = None
    ) -> dict:
        """ Generates JSON schema for keyword extraction.
        """
        json_schema = {
          "type": "object",
          "properties": {
            "topics": {
              "type": "array",
              "items": {
                "type": "string"
              }
            }
          },
          "required": [self.keywords_key]
        }
        if min_keywords:
            json_schema["properties"]["topics"]["minItems"] = min_keywords
        if max_keywords:
            json_schema["properties"]["topics"]["maxItems"] = max_keywords
        if keywords:
            json_schema["properties"][self.keywords_key]["items"]["enum"] = keywords
        return json_schema

    def generate_json_schema(self, **kwargs) -> dict:
        """Generates JSON schema for the given fields.
        """
        if isinstance(kwargs.get("fields")[0], str):
            json_schema = self._generate_json_schema_simple(**kwargs)
        else:
            json_schema = self._generate_json_schema_restricted(**kwargs)
        return json_schema

    def _generate_json_schema_simple(self,
            fields: List[str],
            default_type: str = "string",
            max_length: int = 100,
            min_length: int = 2,
            min_items: int = 0,
            max_items: int = 100,
            required_fields: List[str] = [],
            require_all: bool = False
        ) -> dict:
        """Generates JSON schema for the given fields.
        NB! Currently assumes that the output corresponding to
        each field is a list of `default_type` (string).
        """
        json_schema = {
            "type": "object",
            "properties": {
            field: {
                "type": "array",
                "items": {
                    "type": default_type,
                    "maxLength": max_length,
                    "minLength": min_length
                },
                "minItems": min_items,
                "maxItems": max_items
            }
            for field in fields
          },
        }
        if required_fields:
            json_schema["required"] = required_fields
        elif require_all:
            json_schema["required"] = fields
        return json_schema

    def _generate_json_schema_restricted(self,
            fields: List[dict],
            default_type: str = "string",
            max_length: int = 100,
            min_items: int = 0,
            max_items: int = 10,
            required_fields: List[str] = [],
            require_all: bool = False
        ) -> dict:
        """Generates JSON schema for the given fields according to restrictions
        passed with key `restrictions`.
        `fields` has to be formatted as follows:
        [{"name": <str:field_name>, "restrictions": <dict: restrictions for field>}, ...]
        """

        json_schema = {
          "type": "object",
          "properties": {
            field.get("name"): {
              "type": "array",
              "items": {
                  **{
                    "type": default_type,
                    "maxLength": max_length
                  },
                  **field.get("item_restrictions", {})
              },
              **field.get("list_restrictions", {})
            }
            for field in fields
          }
        }
        if required_fields:
            json_schema["required"] = required_fields
        elif require_all:
            json_schema["required"] = [field.get("name") for field in fields]
        return json_schema



from rara_meta_extractor.epub_meta_extractor import EPUBMetaExtractor
from rara_meta_extractor.tools.utils import jl_generator
from typing import List

import pytest
import os


def load_testset(file_path: str) -> List[dict]:
    testset = []
    for doc in jl_generator(file_path):
        doc_id = doc.get("ester_id")
        data = doc.get("digi_output")
        epub_meta = data.get("doc_meta").get("epub_metadata")
        testset.append(
            {
                "epub_metadata": epub_meta
            }
        )
    return testset

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_FILE = os.path.join(ROOT_DIR, "tests", "test_data", "epub_meta", "digitized_epubs.jl")
TESTSET = load_testset(TEST_FILE)

EPUB_EXTRACTOR = EPUBMetaExtractor(extra_logging={"task_id": 1, "user_id": 1})

def test_epub_extractor():
    for doc_batch in TESTSET:
        meta = EPUB_EXTRACTOR.extract_meta(**doc_batch)
        assert meta



from string import punctuation

# -------------------------------------------- #
# Paterns used by Text Part Classifier
# -------------------------------------------- #

# Table of contents title variants
TABLE_OF_CONTENTS_PATTERNS = [
    r"Sisukord",
    r"Contents",
    r"Table of Contents"
]

# Abstract title variants
ABSTRACT_PATTERNS = [
    r"Abstract",
    r"Annotatsioon"
]
# Conclusion title variants
CONCLUSION_PATTERNS = [
    r"Kokkuvõte(?!:)",
    r"Kokkuvõte ja järeldused",
    r"Kokkuvõte ja soovitused",
    r"Kokkuvõte ja ettepanekud",
    r"Summary",
    r"Summary in Estonian",
    r"Conclusions?",
    r"Lühitutvustus(?!:)",
    r"Lühikokkuvõte(?!:)",
    r"Lühike kokkuvõte(?!:)",
    r"Kokkuvõte auditeerimise tulemustest",
    r"Discussion and conclusions?",
    r"Final Comments and Conclusions?",
    r"Discussion and Conclusion",
    r"The main conclusions of the work are the following",
    r"In Conclusion"
]

ORDINAL_PATTERNS = [
    "first", "second", "third", "fourth", "fifth",
    "esimene", "teine", "kolmas", "neljas", "viies",
    "chapter", "peatükk", "peatüki", "rühmatööst"
]

# -------------------------------------------- #
# SKIP patterns
# -------------------------------------------- #

SKIP_WORDS = ["the", "an", "a", "der", "das"]
SKIP_CHARS = punctuation



from rara_meta_extractor.llama_extractor import LlamaExtractor
from rara_meta_extractor.config import (
    META_EXTRACTOR_CONFIG, TEXT_CLASSIFIER_CONFIG, METADATA_TEXT_BLOCKS
)
from rara_meta_extractor.constants.data_classes import MetaField

class TextClassifierAgent(LlamaExtractor):
    """ Llama agent for classifying texts.
    """
    def __init__(self, config: dict = TEXT_CLASSIFIER_CONFIG):
        default_config: dict = TEXT_CLASSIFIER_CONFIG
        default_config.update(config)
        super().__init__(**default_config)

    def has_meta(self, text: str, timeout: int = 90) -> bool:
        """ Outputs a boolean value indicating, if the
        given text contains metadata or not.
        """
        text_info = self.extract(text, timeout=timeout)
        try:
            text_classes = text_info.get(MetaField.TEXT_TYPE, "")
            if text_classes[0] in METADATA_TEXT_BLOCKS:
                meta = True
            else:
                meta = False
        except:
            meta = False
        return meta


class MetaExtractorAgent(LlamaExtractor):
    """ Llama agent for extracting metadata.
    """
    def __init__(self, config: dict = META_EXTRACTOR_CONFIG):
        default_config: dict = META_EXTRACTOR_CONFIG
        default_config.update(config)
        super().__init__(**default_config)



from rara_meta_extractor.regex_extractors import ISBNRegexExtractor
import pytest


def test_isbn_regex_extraction():
    ex = ISBNRegexExtractor()
    test_texts = [
        "Keskus Praxis ; ISBN 897 -9949 -662-25-5 (pdf)",
        "balabla S 897 9949 662 25 5 aa",
        "balabla ISBN 897     9949 662    25 5 aa",
        "balabla ISBN 897-9949-662-25-5 aa",
        "balabla ISBN 897 9949 662 - 25 5 aa",
        "balabla ISBN 897   -9949-662 - 25- 5 aa"
    ]
    for text in test_texts:
        isbn = ex.extract(text)
        assert len(isbn) == 1
        assert isbn[0] == "8979949662255"


from rara_meta_extractor.epub_meta_extractor import EPUBMetaExtractor
from rara_meta_extractor.text_part_classifiers.epub_text_part_classifier import EPUBTextPartClassifier
from rara_meta_extractor.tools.utils import jl_generator
from typing import List
from pprint import pprint
import pytest
import os

def load_testset(file_path: str) -> List[dict]:
    testset = []
    for doc in jl_generator(file_path):
        doc_id = doc.get("ester_id")
        data = doc.get("digi_output")
        epub_meta = data.get("doc_meta").get("epub_metadata")
        digitized_texts = data.get("texts")
        testset.append(
            {
                "epub_metadata": epub_meta,
                "digitized_texts": digitized_texts
            }
        )
    return testset

def get_classifier_input(file_path: str) -> List[dict]:
    testset = load_testset(file_path)
    extractor = EPUBMetaExtractor()

    classifier_input = []
    for doc_batch in testset:
        epub_meta = doc_batch.get("epub_metadata")
        digitized_texts = doc_batch.get("digitized_texts")
        meta = extractor.extract_meta(epub_metadata=epub_meta)
        pprint(meta)

        classifier_input.append(
            {
                "digitized_texts": digitized_texts,
                "epub_meta": meta
            }
        )
    return classifier_input

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_FILE = os.path.join(ROOT_DIR, "tests", "test_data", "epub_meta", "digitized_epubs.jl")

CLASSIFIER_TESTSET = get_classifier_input(TEST_FILE)
EPUB_CLASSIFIER = EPUBTextPartClassifier(extra_logging={"task_id": 1, "user_id": 1})

#@pytest.mark.skip(reason="EPUB text part classifier currently broken due do changes in EPUB extractor output.")
def test_epub_text_part_classifier():
    # TODO: inspect why @ inices
    no_text_part_indices = {0, 2, 4, 5, 6}
    for i, doc_batch in enumerate(CLASSIFIER_TESTSET):
        epub_text_parts = EPUB_CLASSIFIER.get_parts_of_text(**doc_batch)
        if i not in no_text_part_indices:
            assert epub_text_parts



import statistics
from collections import defaultdict
from copy import deepcopy
from typing import List, NoReturn, Any, Tuple

import regex as re

from rara_meta_extractor.config import (
    LOGGER, META_FIELDS, AUTHOR_FIELDS, META_YEAR_FIELDS, AUTHOR_ROLES_DICT,
    TABLE_OF_CONTENTS_MIN_UPPERCASE_RATIO, TABLE_OF_CONTENTS_MIN_UPPERCASE_AVG_LENGTH,
    AUTHOR_TYPES_MAP, TITLE_TYPES_MAP, META_FIELDS_WITH_MULTIPLE_VALUES,
    DEFAULT_ISSUE_TYPE, META_KEYS_TO_IGNORE, TITLE_KEYS
)
from rara_meta_extractor.constants.data_classes import (
    AuthorField, DataRestrictions, MetaField, AuthorType,
    AuthorNameOrder, TitleType, TitleTypeSimple, TextPartLabel
)
from rara_meta_extractor.constants.patterns import SKIP_CHARS, SKIP_WORDS
from rara_meta_extractor.tools.utils import detect_language

difference = lambda x, y: set(x) - set(y)


class MetaValidator:
    def __init__(self, meta_fields: List[str] = META_FIELDS, author_fields: List[str] = AUTHOR_FIELDS) -> NoReturn:
        self.meta_fields: List[str] = meta_fields
        self.author_fields: List[str] = author_fields

    def _filter_by_length(
            self,
            key: str,
            values: List[Any],
            length: Tuple[int, int],
            values_to_str: bool = False
    ) -> List[Any]:
        """ Filters out values not complying to length requirements.
        """
        original_values = deepcopy(values)
        filtered_values = []
        min_length, max_length = length
        for value in values:
            if min_length <= len(str(value)) <= max_length:
                if values_to_str:
                    filtered_values.append(str(value))
                else:
                    filtered_values.append(value)
        diff = difference(original_values, filtered_values)
        if diff:
            LOGGER.debug(
                f"Removed the following values for key '{key}' not complying " \
                f" to length requirement ({length} characters): {list(diff)}"
            )
        return filtered_values

    def _remove_empty(self, key: str, values: List[Any]) -> List[Any]:
        """ Remove empty strings and "None"-s.
        """
        original_values = deepcopy(values)
        values = [
            v for v in values
            if str(v).strip() and str(v).strip() != "None"
        ]
        n_removed = len(original_values) - len(values)
        if n_removed > 0:
            LOGGER.debug(f"Removed {n_removed} empty values for key '{key}'.")
        return values

    def _strip_punctuation(self, key: str, values: list) -> list:
        """ Remove '-' from ISBN and ISSN.
        """
        if key == MetaField.ISBN or key == MetaField.ISSN:
            stripped_values = []
            for value in values:
                new_value = re.sub(r"-", "", value)
                new_value = re.sub(r"\s+", "", new_value)
                stripped_values.append(new_value)
        else:
            stripped_values = values
        return stripped_values

    def _get_validated_values(self, key: str, values: list,
                              check_dates: bool = True
                              ) -> List[Any]:
        values = self._remove_empty(key, values)
        values = self._strip_punctuation(key, values)

        if key == MetaField.ISBN:
            values = self._filter_by_length(
                key=key,
                values=values,
                length=[DataRestrictions.ISBN_MIN_LENGTH, DataRestrictions.ISBN_MAX_LENGTH],
                values_to_str=True
            )
        elif key == MetaField.ISSN:
            values = self._filter_by_length(
                key=key,
                values=values,
                length=[DataRestrictions.ISSN_LENGTH, DataRestrictions.ISSN_LENGTH],
                values_to_str=True
            )
        elif check_dates and key.strip() in META_YEAR_FIELDS:
            values = self._filter_by_length(
                key=key,
                values=values,
                length=[DataRestrictions.YEAR_LENGTH, DataRestrictions.YEAR_LENGTH],
                values_to_str=False
            )
        return values

    def _is_valid_key(self, key: str) -> bool:
        """ Checks, if the key is a valid key present in either
        META_FIELDS or AUTHOR_FIELDS list.
        """
        split_key = " ".join(key.split("_"))
        if key in META_FIELDS or split_key in META_FIELDS:
            return True
        if key in AUTHOR_FIELDS or split_key in AUTHOR_FIELDS:
            return True
        return False


class MetaFormatter:
    """ Formats meta
    """

    def __init__(self,
                 meta_fields: List[str] = META_FIELDS,
                 author_fields: List[str] = AUTHOR_FIELDS,
                 extra_logging: dict = None,
                 ) -> NoReturn:
        self.extra_logging = extra_logging or {}
        self.meta_validator = MetaValidator(
            meta_fields=meta_fields,
            author_fields=author_fields
        )

    def _filter_by_ratio(self, values: List[Tuple[Any, int]],
                         n_trials: int, min_ratio: float
                         ) -> List[Any]:
        """ Keeps only values that have been predicted
        in `min_ratio` trials.
        """
        filtered_values = [
            v[0] for v in values
            if float(v[1] / n_trials) >= min_ratio
        ]
        return filtered_values

    def _merge_meta(self, meta_batches: List[dict], min_ratio: float = 0.5) -> dict:
        """ Merges meta into a single dict.
        """
        LOGGER.debug("Merging and formatting metadata.", extra=self.extra_logging)

        formatted_meta = [
            self._format_meta(meta_dict)
            for meta_dict in meta_batches
            if meta_dict
        ]
        meta = {}
        frequencies = defaultdict(lambda: defaultdict(int))
        n_trials = len(formatted_meta)

        for meta_dict in formatted_meta:
            for key, values in meta_dict.items():
                if key == MetaField.TITLES:
                    # "titles" list contains only the main title
                    # for the latest model, so merge them into one
                    values = [" ".join(values)]
                for value in values:
                    if not isinstance(value, str):
                        LOGGER.debug(f"Unhashable value: '{value}'. Skipping.", extra=self.extra_logging)
                        continue
                    frequencies[key][value] += 1

        for key, value_dict in frequencies.items():
            value_list = sorted(
                list(value_dict.items()),
                key=lambda x: x[1],
                reverse=True
            )
            meta[key] = self._filter_by_ratio(
                values=value_list,
                n_trials=n_trials,
                min_ratio=min_ratio
            )
        return meta

    def _parse_authors(self, authors_list: List[str]) -> List[str]:
        """ Parse weirdly extracted authors like:
        "Reelika RätsepMari-Liis TammikElmar Zimmer"
        """
        LOGGER.debug("Parsing authors.")
        new_authors = []
        for author in authors_list:
            parsed = re.split("(?<=[a-züõöäšž])(?=[A-ZÜÕÖÄŠŽ])", author)
            new_authors.extend(parsed)
        return new_authors

    def _parse_dates(self, values: List[str]) -> List[str]:
        parsed = []
        for value in values:
            match = re.search(r"(?<=\D|^)\d{4}(?=\D|$)", value)
            if match:
                parsed.append(match.group())
        return parsed

    def _add_missing_keys(self, meta: dict,
                          keys_to_ignore: List[str] = META_KEYS_TO_IGNORE
                          ) -> dict:
        """ Adds missing meta keys.
        """
        meta_copy = deepcopy(meta)
        missing_keys = {}
        for key in META_FIELDS:
            if key not in meta and key not in keys_to_ignore:
                missing_keys[key] = []
        meta_copy.update(missing_keys)
        return meta_copy

    def _remove_empty_fields(self, meta: dict) -> dict:
        """ Removes empty fields.
        """
        new_meta = {}
        for key, values in meta.items():
            if values:
                new_meta[key] = values
        return new_meta

    def _format_meta(self, meta: dict, check_dates: bool = False) -> dict:
        """ Format meta.
        """
        formatted_meta = {}
        for key, values in list(meta.items()):
            key = key.strip()
            if not self.meta_validator._is_valid_key(key):
                LOGGER.debug(
                    f"Detected an invalid key '{key}'. " \
                    f"This will NOT be added to the output.",
                    extra=self.extra_logging
                )
                continue
            values = self.meta_validator._get_validated_values(
                key=key, values=values, check_dates=check_dates
            )
            if values:
                if key in AUTHOR_FIELDS:
                    formatted_meta[key] = self._parse_authors(values)
                elif key in META_YEAR_FIELDS:
                    formatted_meta[key] = self._parse_dates(values)
                else:
                    formatted_meta[key] = values
        return formatted_meta

    def _format_authors(self, meta: dict, simple: bool = False) -> dict:
        """ Convert authors from a flat structure into
        a list of dicts.
        """
        LOGGER.debug("Formatting authors.", extra=self.extra_logging)
        new_meta = {MetaField.AUTHORS: []}  # , "main_author": ""}
        for key, values in list(meta.items()):
            if key in AUTHOR_FIELDS:
                for i, value in enumerate(values):
                    if i == 0 and key == AuthorField.AUTHOR:
                        is_primary = True
                    else:
                        is_primary = False
                    new_author = Author(
                        name=value,
                        role=key,
                        is_primary=is_primary,
                        extra_logging=self.extra_logging
                    ).to_dict(simple=simple)
                    new_meta["authors"].append(new_author)
            else:
                new_meta[key] = values
        return new_meta


class Author:
    def __init__(self,
                 name: str,
                 role: str,
                 is_primary: bool,
                 author_type: str = AuthorType.UNK,
                 unknown_role: str = AuthorField.UNKNOWN,
                 map_role: bool = True,
                 author_name_order: int = AuthorNameOrder.FIRST_NAME_FIRST,
                 extra_logging: dict = None
                 ):
        self.extra_logging = extra_logging or {}
        self.original_name: str = name
        self.is_primary: bool = is_primary
        self.en_role: str = role
        self.et_role: str = AUTHOR_ROLES_DICT.get(self.en_role, unknown_role) if map_role else role
        self.author_type: str = AUTHOR_TYPES_MAP.get(author_type, AuthorType.UNK)
        self.author_name_order: int = author_name_order

    @property
    def name(self):
        try:
            if "," in self.original_name.strip(","):
                name_tokens = self.original_name.split(",")
                name = f"{name_tokens[1].strip()} {name_tokens[0].strip()}"
            else:
                name = self.original_name
        except Exception as e:
            LOGGER.error(f"{e}. original_name={self.original_name}", extra=self.extra_logging)
            name = self.original_name
        return name

    @staticmethod
    def is_false_positive(author_name: str, max_n_words: int = 7,
                          max_length: int = 50
                          ) -> bool:
        """ Author names detected from EPUBs might sometimes
        contain just a long free form text like:
        'E-raamatu teostus Ants Tuur / www.flagella.ee '
        '- Kvaliteetsed eraamatu lahendused.'
        """
        try:
            name_tokens = [
                token.strip()
                for token in author_name.split()
                if token.strip()
            ]
            if len(name_tokens) > max_n_words or len(author_name) > max_length:
                return True
            if re.search(r"http(s)?[:]", author_name):
                return True
        except Exception as e:
            LOGGER.error(
                f"Failed parsing author (author_name='{author_name}'. Returing it as false positive. Exception: {e}.")
            return True
        return False

    def to_dict(self, simple: bool = False) -> dict:
        if simple:
            author_dict = {
                "name": self.name,
                "role": self.et_role
            }
        else:
            author_dict = {
                "is_primary": self.is_primary,
                "name": self.name,
                "role": self.et_role,
                "type": self.author_type,
                "name_order": self.author_name_order
            }
        return author_dict


class Title:
    def __init__(self, titles: List[str] | dict, lang: str = "",
                 part_number: str = "", title_type: str = ""
                 ):
        self.titles: List[str] | dict = titles
        self.lang: str = lang
        self.title_type: str = title_type
        self.__part_number: str = part_number

    @property
    def part_number(self) -> str:
        if not self.__part_number:
            if isinstance(self.titles, dict):
                part_numbers = self.titles.get(MetaField.TITLE_PART_NR, [])
                self.__part_number = part_numbers[0] if part_numbers else ""
        return self.__part_number

    @staticmethod
    def find_skip(title: str) -> int:
        """ Number of chars that should not be included into search,
        e.g. Articles like 'the', 'an', various punctuation chars etc.
        """
        skip = 0
        skip_pattern = rf"^(([{SKIP_CHARS}]|({'|'.join(SKIP_WORDS)})\s)\s*)+(?=\S)"
        match = re.search(skip_pattern, title, re.IGNORECASE)
        if match:
            skip = len(match.group())
        return skip

    def _get_title_type(self, i: int | None = None, key: str | None = None,
                        simple: bool = False, from_list: bool = True
                        ) -> str:
        if self.title_type:
            return self.title_type
        if simple:
            title_type_class = TitleTypeSimple
        else:
            title_type_class = TitleType

        if from_list:
            title_type_map = {
                0: title_type_class.TITLE,
                1: title_type_class.ADDITIONAL_TITLE,
                2: title_type_class.PARALLEL_TITLE
            }
            title_type = title_type_map.get(i, "")
        else:
            title_type_map = {
                MetaField.TITLE: title_type_class.TITLE,
                MetaField.SUBTITLE: title_type_class.ADDITIONAL_TITLE,
                MetaField.TITLE_PART_NAME: title_type_class.PARALLEL_TITLE

            }
            title_type = title_type_map.get(key, "")
        return title_type

    def _clean(self, title: str) -> str:
        title = title.strip()
        title = title.strip("/")
        title = title.strip()
        title = title.strip(":")
        title = title.strip()
        return title

    def to_dicts_from_list(self, simple: bool = False) -> List[dict]:
        formatted = []
        for i, title in enumerate(self.titles):
            if i > 2 and not simple:
                break

            title_type = self._get_title_type(i=i, simple=simple, from_list=True)
            title_type_int = TITLE_TYPES_MAP.get(title_type, None)
            skip = Title.find_skip(title)
            title_language = detect_language(title) if not self.lang else self.lang

            if not simple:
                if title_type == TitleType.ADDITIONAL_TITLE:
                    title_dict = formatted.pop()
                    title_dict["part_title"] = title
                else:
                    title_dict = {
                        "title": self._clean(title),
                        "title_language": title_language,
                        "part_number": self.part_number,
                        "part_title": "",
                        "version": "",
                        "author_from_title": "",
                        "skip": skip,
                        "title_type": title_type,
                        "title_type_int": title_type_int
                    }
            else:
                title_dict = {
                    "title": title,
                    "title_type": title_type
                }
            formatted.append(title_dict)
        return formatted

    def to_dicts_from_dict(self, simple: bool = False) -> List[dict]:
        formatted = []
        keys_to_ignore = [MetaField.TITLE_PART_NR, MetaField.SUBTITLE, MetaField.TITLE_VARFORM]

        for key, titles in self.titles.items():
            if key in keys_to_ignore or not titles:
                continue
            title = " ".join(titles)
            title_type = self._get_title_type(key=key, simple=simple, from_list=False)
            title_type_int = TITLE_TYPES_MAP.get(title_type, None)
            skip = Title.find_skip(title)
            title_language = detect_language(title) if not self.lang else self.lang
            subtitles = self.titles.get(MetaField.SUBTITLE, [])
            subtitle = subtitles[0] if subtitles else ""

            if not simple:
                if key == MetaField.TITLE:
                    title_dict = {
                        "title": self._clean(title),
                        "title_language": title_language,
                        "part_number": self.part_number,
                        "part_title": self._clean(subtitle),
                        "version": "",
                        "author_from_title": "",
                        "skip": skip,
                        "title_type": title_type,
                        "title_type_int": title_type_int
                    }
                else:
                    title_dict = {
                        "title": self._clean(title),
                        "title_language": title_language,
                        "part_number": self.part_number,
                        "part_title": "",
                        "version": "",
                        "author_from_title": "",
                        "skip": skip,
                        "title_type": title_type,
                        "title_type_int": title_type_int
                    }
            else:
                title_dict = {
                    "title": title,
                    "title_type": title_type
                }
            formatted.append(title_dict)
        return formatted

    def to_dicts(self, simple: bool = False) -> List[dict]:
        if isinstance(self.titles, dict):
            formatted = self.to_dicts_from_dict(simple)
        else:
            formatted = self.to_dicts_from_list(simple)
        return formatted


class TextPart:
    def __init__(self, text_type: str, text_value: str, language: str = ""):
        self.text_type = text_type
        self.text_value = text_value
        self.language = language if language else detect_language(text_value)

    def to_dict(self) -> dict:
        text_part_dict = {
            "text_type": self.text_type,
            "text_value": self.text_value,
            "language": self.language
        }
        return text_part_dict


class Meta:
    def __init__(
            self,
            meta_batches: List[dict],
            text_parts: List[dict],
            min_ratio: float = 0.8,
            add_missing_keys: bool = False,
            simple: bool = False,
            language: str = "",
            extra_logging: dict = None
    ) -> NoReturn:
        self.meta_batches: List[dict] = meta_batches
        self.text_parts: List[dict] = text_parts
        self.min_ratio: float = min_ratio
        self.add_missing_keys: bool = add_missing_keys
        self.simple: bool = simple
        self.language: str = language
        self.extra_logging: dict = extra_logging or {}

        self.meta_formatter: MetaFormatter = MetaFormatter(
            meta_fields=META_FIELDS,
            author_fields=AUTHOR_FIELDS,
            extra_logging=self.extra_logging
        )

        self.__merged_meta: dict = {}
        self.__meta_with_reformatted_authors: dict = {}
        self.__meta_with_all_keys: dict = {}
        self.__meta_without_empty_fields: dict = {}

    @property
    def merged_meta(self) -> dict:
        if not self.__merged_meta:
            self.__merged_meta = self.meta_formatter._merge_meta(
                meta_batches=self.meta_batches,
                min_ratio=self.min_ratio
            )
        return self.__merged_meta

    @property
    def meta_with_reformatted_authors(self):
        if not self.__meta_with_reformatted_authors:
            self.__meta_with_reformatted_authors = self.meta_formatter._format_authors(
                meta=self.merged_meta, simple=self.simple
            )
        return self.__meta_with_reformatted_authors

    @property
    def meta_with_all_keys(self):
        if not self.__meta_with_all_keys:
            self.__meta_with_all_keys = self.meta_formatter._add_missing_keys(
                meta=self.meta_without_empty_fields
            )
        return self.__meta_with_all_keys

    @property
    def meta_without_empty_fields(self):
        if not self.__meta_without_empty_fields:
            self.__meta_without_empty_fields = self.meta_formatter._remove_empty_fields(
                meta=self.meta_with_reformatted_authors
            )
        return self.__meta_without_empty_fields

    def _concat(self, meta: dict) -> dict:
        concatted = {}
        for key, value in meta.items():
            key_tokens = key.split()
            new_key = "_".join(key_tokens)
            concatted[new_key] = value
        return concatted

    @staticmethod
    def update_field_types(meta_dict: dict, empty_value: Any = "", custom_keys_to_ignore: List[str] = []) -> dict:
        keys_to_ignore = [MetaField.TITLES, MetaField.AUTHORS, MetaField.TEXT_PARTS]
        keys_to_ignore.extend(META_FIELDS_WITH_MULTIPLE_VALUES)
        keys_to_ignore.extend(custom_keys_to_ignore)
        for key, values in list(meta_dict.items()):
            if key not in keys_to_ignore:
                if not isinstance(values, list):
                    continue
                if values:
                    meta_dict[key] = values[0]
                else:
                    meta_dict[key] = empty_value
        return meta_dict

    def _remove_publisher_from_authors(self, meta_dict: dict) -> dict:
        LOGGER.debug(f"Removing publisher from authors...", extra=self.extra_logging)
        authors = meta_dict.pop(MetaField.AUTHORS, [])

        filtered_authors = []
        for author in authors:
            if author.get("role").lower() == "väljaandja":
                LOGGER.debug(f"Detected publisher in authors list. Removing it.", extra=self.extra_logging)
                meta_dict[MetaField.PUBLISHER] = author.get("name")
            else:
                filtered_authors.append(author)

        meta_dict[MetaField.AUTHORS] = filtered_authors
        return meta_dict

    def to_dict(self):
        if self.add_missing_keys:
            meta_dict = self._concat(self.meta_with_all_keys)
        else:
            meta_dict = self._concat(self.meta_without_empty_fields)
        if self.text_parts or self.add_missing_keys:
            meta_dict[MetaField.TEXT_PARTS] = self.text_parts
        if MetaField.TITLES in meta_dict:
            titles = meta_dict.pop(MetaField.TITLES)
            formatted_titles = Title(titles=titles, lang=self.language).to_dicts(simple=self.simple)
            meta_dict[MetaField.TITLES] = formatted_titles
        elif MetaField.TITLE in meta_dict:
            titles = {}
            for key in TITLE_KEYS:
                value = meta_dict.pop(key, "")
                if not value:
                    value = meta_dict.pop("_".join(key.split()), "")
                titles[key] = value
            formatted_titles = Title(titles=titles, lang=self.language).to_dicts(simple=self.simple)
            meta_dict[MetaField.TITLES] = formatted_titles

        meta_dict = Meta.update_field_types(meta_dict)
        # Hotfix
        keys_to_remove = ["country_from_008", "unknown"]
        new_dict = {}
        for key, values in meta_dict.items():
            if key not in keys_to_remove:
                new_dict[key] = values
        meta_dict = new_dict

        # Always add issue type (bib style):
        if MetaField.ISSUE_TYPE not in meta_dict:
            meta_dict[MetaField.ISSUE_TYPE] = DEFAULT_ISSUE_TYPE

        meta_dict = self._remove_publisher_from_authors(meta_dict)

        return meta_dict


class TableOfContents:
    def __init__(self, extracted_meta: dict, extra_logging: dict = None):
        self.extra_logging = extra_logging or {}

        self.__chapters: List[str] = []
        self.__table_of_contents_str: str = ""
        self.__table_of_contents_lang: str = ""
        self._text_parts: List[dict] = extracted_meta.get(MetaField.TEXT_PARTS, [])
        self._authors: List[dict] = extracted_meta.get(MetaField.AUTHORS, [])
        self._sections: List[dict] = extracted_meta.get(MetaField.SECTIONS, [])

    def _clean_chapter(self, chapter: str) -> str:
        LOGGER.debug(f"Chapter title before cleaning: '{chapter}'", extra=self.extra_logging)
        chapter = re.sub(r"(?<=[^.])\s*([.]\s*)+(\s*\d*|[IVXLCDM]*)*(?=$)", "", chapter)
        LOGGER.debug(f"Chapter title after cleaning: '{chapter}'", extra=self.extra_logging)
        return chapter

    def _split_chapter(self, cleaned_chapter: str) -> List[str]:
        split_pattern = r"([.]\s*){3,}((\s*\d*|[IVXLCDM]*)\s)*"
        split_match = re.search(split_pattern, cleaned_chapter)
        chapters = [cleaned_chapter]
        if split_match:
            LOGGER.debug(
                f"Detected a split pattern in chapter '{cleaned_chapter}'. Splitting it into two.",
                extra=self.extra_logging
            )
            splitter = split_match.group()
            chapters = cleaned_chapter.split(splitter, 1)
            chapters = [c.strip() for c in chapters]
        return chapters

    def _filter_by_case(
            self,
            chapters: List[str],
            min_ratio: float = TABLE_OF_CONTENTS_MIN_UPPERCASE_RATIO,
            min_avg_length: int = TABLE_OF_CONTENTS_MIN_UPPERCASE_AVG_LENGTH
    ) -> List[str]:
        """ If some titles are in uppercase and others are not,
        assume that the lower-case titles are false positives
        and filter them out.
        """
        chapter_stats = {
            "upper": [],
            "other": [],
            "upper_lens": [],
            "other_lens": []
        }
        out = chapters
        for chapter in chapters:
            if chapter.upper() == chapter:
                chapter_stats["upper"].append(chapter)
                chapter_stats["upper_lens"].append(len(chapter))
            else:
                chapter_stats["other"].append(chapter)
                chapter_stats["other_lens"].append(len(chapter))
        if chapters and len(chapter_stats["upper"]) / len(chapters) >= min_ratio:
            if statistics.mean(chapter_stats["upper_lens"]) >= min_avg_length:
                out = chapter_stats["upper"]
        return out

    @property
    def table_of_contents_lang(self) -> str:
        if not self.__table_of_contents_lang:
            for text_part in self._text_parts:
                text_type = text_part.get("text_type", "")
                if text_type == TextPartLabel.TABLE_OF_CONTENTS:
                    self.__table_of_contents_lang = text_part.get("language", "")
        return self.__table_of_contents_lang

    @property
    def table_of_contents_str(self) -> str:
        if not self.__table_of_contents_str:
            for text_part in self._text_parts:
                text_type = text_part.get("text_type", "")
                if text_type == TextPartLabel.TABLE_OF_CONTENTS:
                    self.__table_of_contents_str = text_part.get("text_value", "")
        return self.__table_of_contents_str

    @property
    def table_of_contents(self) -> List[str]:
        """ Extract chapter names from table of contents.
        """
        if not self.__chapters:
            if self._sections:
                chapters = []
                languages = defaultdict(int)
                most_frequent_lang = ""
                for section in self._sections:
                    titles = section.get(MetaField.TITLES, [])
                    _authors = section.get(MetaField.AUTHORS, [])
                    lang = section.get("language", "")

                    languages[lang] += 1
                    chapter = titles[0].get("title") if titles else ""
                    authors = [author.get("name") for author in _authors]
                    if chapter:
                        chapters.append(
                            {"chapter": chapter, "authors": authors}
                        )
                if languages:
                    most_frequent_lang = sorted(
                        list(languages.items()),
                        key=lambda x: x[1],
                        reverse=True
                    )[0][0]
                self.__chapters = {
                    "content": chapters,
                    "language": most_frequent_lang
                }

            else:
                raw_chapters = self.table_of_contents_str.split("\n")
                _chapters = [
                    chapter.strip()
                    for chapter in raw_chapters
                    if chapter.strip()
                ]
                filtered_chapters = self._filter_by_case(_chapters)
                content = []
                for chapter in filtered_chapters:
                    cleaned_chapter = self._clean_chapter(chapter)
                    chapters = self._split_chapter(cleaned_chapter)
                    for chapter in chapters:
                        content.append({"chapter": chapter, "authors": []})
                self.__chapters = {
                    "content": content,
                    "language": self.table_of_contents_lang

                }
        return self.__chapters



from rara_meta_extractor.text_part_classifiers.regex_text_part_classifier import RegexTextPartClassifier
from rara_meta_extractor.config import TextPartLabel

import pytest
import os

def load_text(file_path: str) -> str:
    with open(file_path, "r") as f:
        text = f.read()
    return text

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_DATA_DIR = os.path.join(ROOT_DIR, "tests", "test_data", "text_part", "regex")

TEST_TEXTS = {
    TextPartLabel.CONCLUSION: load_text(os.path.join(TEST_DATA_DIR, "conclusion.txt")),
    TextPartLabel.ABSTRACT: load_text(os.path.join(TEST_DATA_DIR, "abstract.txt")),
    TextPartLabel.TABLE_OF_CONTENTS: load_text(os.path.join(TEST_DATA_DIR, "table_of_contents.txt")),
    TextPartLabel.OTHER: load_text(os.path.join(TEST_DATA_DIR, "other.txt"))
}

def test_text_part_classifier():
    tpc = RegexTextPartClassifier(extra_logging={"task_id": 1, "user_id": 1})
    for true_label, test_text in TEST_TEXTS.items():
        predicted_label = tpc.get_label(test_text)
        assert true_label == predicted_label



import regex as re
from typing import List
from rara_meta_extractor.config import LOGGER

class ISBNRegexExtractor:
    """ Regex-based ISBN extractor.
    """
    def __init__(self):
        self.pattern = re.compile(r"(?<=\D|^)(\d\s*-?\s*){13}(?=\D|[^-]|$)")

    def _clean(self, isbn: str) -> str:
        """ Cleans extracted ISBNs from "-" and from whitespaces.

        Parameters
        -----------
        isbn: str
            Extracted raw ISBN number which might contain additional
            punctuation characters.

        Returns
        -----------
        str:
            Cleaned ISBN number.
        """
        isbn = re.sub(r"-", "", isbn)
        isbn = re.sub(r"\s+", "", isbn)
        isbn = isbn.strip()
        return isbn

    def _clean_isbns(self, isbns: List[str]) -> List[str]:
        """ Clean ISBN matches

        Parameters
        -----------
        isbns: List[str]
            ISBN matches to clean.

        Returns
        ----------
        List[str]:
            Cleaned ISBN numbers.
        """
        cleaned = [self._clean(isbn) for isbn in isbns]
        return cleaned


    def _filter_isbns(self, isbns: List[str]) -> List[str]:
        """ Filter out matches that contain "\n" as they
        are very probable false positives. ISBN numbers do not usually
        span over multiple lines.

        Parameters
        -----------
        isbns: List[str]
            ISBN matches to filter.

        Returns
        ----------
        List[str]:
            Filtered ISBN numbers.
        """
        filtered = []
        for isbn in isbns:
            if "\n" not in isbn.strip():
                filtered.append(isbn)
        return filtered

    def extract(self, text: str) -> List[str]:
        """ Extracts ISBN numbers from the input text.

        Parameters
        -----------
        text: str
            Text from where to extract the numbers.

        Returns
        ----------
        List[str]:
            Extracted and cleaned ISBN numbers.
        """
        isbns = [match.group() for match in re.finditer(self.pattern, text)]

        isbns = self._filter_isbns(isbns)
        isbns = self._clean_isbns(isbns)
        return isbns



from collections import defaultdict
from copy import deepcopy
from typing import List

import regex as re

from rara_meta_extractor.config import (
    META_EXTRACTOR_CONFIG, TEXT_CLASSIFIER_CONFIG, METADATA_TEXT_BLOCKS, LOGGER,
    MAX_LENGTH_PER_TEXT, DEFAULT_EPUB_MERGE_METHOD, DEFAULT_LLAMA_MERGE_METHOD,
    EPUB_FIELD_METHOD_MAP, LLAMA_FIELD_METHOD_MAP, DEFAULT_ISSUE_TYPE
)
from rara_meta_extractor.constants.data_classes import (
    ExtractorType, MetaField, TextBlock, TextPartLabel, HostEntryField,
    IssueType, SeriesField
)
from rara_meta_extractor.epub_meta_extractor import EPUBMetaExtractor
from rara_meta_extractor.llm_agents import TextClassifierAgent, MetaExtractorAgent
from rara_meta_extractor.mets_alto_meta_extractor import MetsAltoMetaExtractor
from rara_meta_extractor.text_part_classifiers.epub_text_part_classifier import EPUBTextPartClassifier
from rara_meta_extractor.text_part_classifiers.mets_alto_text_part_classifier import MetsAltoTextPartClassifier
from rara_meta_extractor.text_part_classifiers.regex_text_part_classifier import RegexTextPartClassifier
from rara_meta_extractor.tools.epub_llama_merger import EpubLlamaMerger
from rara_meta_extractor.tools.llama_validator import LlamaValidator
from rara_meta_extractor.tools.meta_formatter import Meta, TextPart, TableOfContents
from rara_meta_extractor.tools.utils import detect_language


class MetaExtractor:
    def __init__(
            self,
            meta_extractor_config: dict = META_EXTRACTOR_CONFIG,
            text_classifier_config: dict = TEXT_CLASSIFIER_CONFIG,
            extra_logging: dict = None,
    ):
        self.extra_logging = extra_logging or {}
        self.meta_extractor_config = meta_extractor_config
        self.meta_extractor_config["extra_logging"] = self.extra_logging
        self.text_classifier_config = text_classifier_config
        self.text_classifier_config["extra_logging"] = self.extra_logging

        self.meta_agent = MetaExtractorAgent(self.meta_extractor_config)
        self.text_classifier_agent = TextClassifierAgent(self.text_classifier_config)

        self.text_part_classifier = RegexTextPartClassifier(extra_logging=self.extra_logging)
        self.epub_text_part_classifier = EPUBTextPartClassifier(extra_logging=self.extra_logging)
        self.mets_alto_text_part_classifier = MetsAltoTextPartClassifier(extra_logging=self.extra_logging)

        self.epub_meta_extractor = EPUBMetaExtractor(extra_logging=self.extra_logging)

        self.llama_validator = LlamaValidator(extra_logging=self.extra_logging)

    def classify_text(self, text: str, default: str = TextBlock.METADATA) -> str:
        """ Classifies text into one of the text blocks defined in TextBlock.

        Parameters
        -----------
        text: str
            Text to classify
        default: str
            Default value to return in case of an exception.

        Returns
        -----------
        str:
            Text class.
        """
        try:
            text_type_dict = self.text_classifier_agent.extract(text=text)
            LOGGER.debug(f"Llama text classifier output: {text_type_dict}", extra=self.extra_logging)
            text_class = text_type_dict.get("text_type")[0]
        except Exception as e:
            LOGGER.exception(
                f"Detecting text type for text '{text[:50]}' failed with error: {e}. "
                f"Defaulting to '{default}'",
                extra=self.extra_logging
            )
            text_class = default
        return text_class

    def _construct_llama_input(
            self, texts: List[str], use_llm: bool = False,
            max_length_per_text: int = 1500, n_first_pages: int = 5,
            n_last_pages: int = 0, n_strict_include: int = 3
    ) -> List[str]:
        """ Construct input for Llama.

        Parameters
        -----------
        texts: List[str]
            Raw texts
        use_llm: bool
            If enabled, LLM is used for each n_first_page and n_last_page text
            to detect, if they contain metadata or not. Only texts containing
            metadata will be passed to Llama for metadata extraction.
        max_length_per_text: int
            Used only if `use_llm=False` and is not applied to `n_strict_include`
            first texts. For other texts, this param will be used to decide, if
            the text will be added to Llama input or not (texts longer than the
            threshold set with this param will be ignored).
        n_first_pages: int
            How many first pages to consider as potential inputs?
        n_last_pages: int
            How many last pages to consider as potential inputs?
        n_strict_include: int
            How many first pages to always include?

        Returns
        -----------
        List[str]:
            List of verified texts
        """
        LOGGER.debug("Constructing Llama input!", extra=self.extra_logging)
        n_pages = len(texts)

        # If number of pages is smaller than number of
        # first and last pages combined, modify the
        # number of last pages in order to avoid
        # duplicates in final Llama input
        if n_pages < n_first_pages + n_last_pages:
            n_last_pages = max(n_pages - n_first_pages, 0)

        first_pages = texts[:n_first_pages]
        last_pages = texts[-n_last_pages:] if n_last_pages > 0 else []
        if n_strict_include > n_first_pages:
            n_strict_include = n_first_pages
        texts_to_analyze = first_pages + last_pages

        if use_llm:
            verified_texts = []
            for text in texts_to_analyze:
                text_type = self.classify_text(text)
                if text_type in METADATA_TEXT_BLOCKS:
                    verified_texts.append(text)
            if not verified_texts:
                LOGGER.error(
                    f"No verified metadata text block found from texts {texts} with LLM!",
                    extra=self.extra_logging
                )
        else:
            verified_texts = texts_to_analyze[:n_strict_include]

            for text in texts_to_analyze[n_strict_include:]:
                if (len(text) < max_length_per_text and not
                re.search("sisukord|table of contents", text, re.IGNORECASE)
                ):
                    verified_texts.append(text)

        return verified_texts

    def _get_text_parts(self, texts: List[dict]) -> List[dict]:
        """ Wrapper function for detecting text partsself.

        Parameters
        -----------
        texts: List[dict]
            Raw texts.

        Returns
        -----------
        List[dict]:
            List of dicts where each dict contains
            one text part label, corresponding text block and
            language of the text block. NB! multiple texts might be concatted
            into one, if conclusion spans over multiple pages etc.

        """
        text_parts = []
        _text_parts = defaultdict(lambda: defaultdict(list))

        for i, text in enumerate(texts):
            LOGGER.debug(f"Classifying text number {i + 1}...", extra=self.extra_logging)
            raw_text = text.get("text", "")
            lang = text.get("language", "")
            label = self.text_part_classifier.get_label(raw_text)
            if label != TextPartLabel.OTHER:
                _text_parts[label]["texts"].append(raw_text)
                _text_parts[label]["langs"].append(lang)

        for label, values in _text_parts.items():
            texts = values.get("texts")
            langs = values.get("langs")
            lang_counts = defaultdict(int)
            for lang in langs:
                lang_counts[lang] += 1
            most_frequent_lang = sorted(
                list(lang_counts.items()),
                key=lambda x: x[1],
                reverse=True
            )[0][0]
            text_part = "\n".join(texts)
            if not most_frequent_lang:
                most_frequent_lang = detect_language(text_part)
            new_text_part = TextPart(
                text_type=label,
                text_value=text_part,
                language=most_frequent_lang
            ).to_dict()

            text_parts.append(new_text_part)
        return text_parts

    def _get_host_entry(self, meta: dict) -> dict:
        LOGGER.debug(f"Constructing host_entry.", extra=self.extra_logging)
        # Meta field has a lot of fields with whitespaces, we need
        # them with underscore : S
        pub_date_field = "_".join(MetaField.PUBLICATION_DATE.split())
        issue_type = meta.get(MetaField.ISSUE_TYPE, "")
        publication_date = meta.get(pub_date_field, "")
        host_entry = {
            HostEntryField.PUBLICATION_DATE: publication_date,
            HostEntryField.NAME: "",
            HostEntryField.PART_NUMBER: ""
        }
        if issue_type in [IssueType.JOURNAL, IssueType.PERIODICAL, IssueType.NEWSPAPER]:
            titles = meta.get("titles", [])
            if titles:
                series_name = titles[0]["title"]
                part_number = titles[0]["part_number"]
                host_entry[HostEntryField.NAME] = series_name
                host_entry[HostEntryField.PART_NUMBER] = part_number
        if not host_entry.get(HostEntryField.NAME, ""):
            series_name = meta.get(MetaField.SERIES_NAME, "")
            if series_name:
                host_entry[HostEntryField.NAME] = series_name
        if not host_entry.get(HostEntryField.PART_NUMBER, ""):
            series_volume = meta.get(MetaField.SERIES_VOLUME, "")
            if series_volume:
                host_entry[HostEntryField.PART_NUMBER] = series_volume
        # if only publication date is filled, remove it from the host entry
        if not host_entry.get(HostEntryField.NAME, None) and not host_entry.get(HostEntryField.PART_NUMBER, None):
            host_entry[HostEntryField.PUBLICATION_DATE] = ""

        return host_entry

    def _get_series_info(self, meta: dict) -> dict:
        LOGGER.debug(f"Constructing series info.", extra=self.extra_logging)
        series_info = {
            SeriesField.ISSN: meta.get(MetaField.SERIES_ISSN, ""),
            SeriesField.STATEMENT: meta.get(MetaField.SERIES_NAME, ""),
            SeriesField.VOLUME: meta.get(MetaField.SERIES_VOLUME, "")
        }
        return series_info

    def extract_from_digitizer_output(self, digitizer_output: dict, **kwargs) -> dict:
        """ Extracts metadata from Digitizer output.
        """
        LOGGER.info("Extracting metadata from Digitizer output.", extra=self.extra_logging)

        texts = digitizer_output.get("texts", [])
        doc_meta = digitizer_output.get("doc_meta", {})

        mets_alto_metadata = doc_meta.get("mets_alto_metadata", [])
        epub_metadata = doc_meta.get("epub_metadata", {})
        languages = doc_meta.get("languages", [])
        language = ""
        if languages:
            language = languages[0].get("language", "")

        result = self.extract(
            texts=texts,
            epub_metadata=epub_metadata,
            mets_alto_metadata=mets_alto_metadata,
            language=language,
            **kwargs
        )
        return result

    def extract_simple(
            self, text: str, lang: str,
            simple: bool = True, **kwargs
    ) -> dict:
        """ For simple usage. Passes plaintext to Llama extractor.
        """
        texts = [{"text": text, "lang": lang}]
        result = self.extract(texts=texts, simple=simple, **kwargs)
        return result

    def extract(
            self,
            texts: List[dict],
            epub_metadata: dict = {},
            mets_alto_metadata: List[str] = [],
            verify_texts_with_llm: bool = False,
            n_trials: int = 1,
            merge_texts: bool = True,
            min_ratio: float = 0.8,
            add_missing_keys: bool = False,
            detect_text_parts: bool = True,
            max_length_per_text: int = MAX_LENGTH_PER_TEXT,
            n_first_pages: int = 5,
            n_last_pages: int = 0,
            n_strict_include: int = 3,
            simple: bool = False,
            llama_api_timeout: int = 90,
            language: str = "",
            validate_llama_output: bool = True,
            allowed_llama_exceptions=(),
            verify_llama_request=False,
    ) -> dict:
        """ Extracts relevant metadata from a batch of texts

        Parameters
        -----------
        texts: List[dict]
            List of texts from where to extract meta information. For EPUB and METS/ALTO,
            expects content of `texts` from digitizer output. Otherwise, must minimally
            contain keys `text` and `lang`.
        epub_metadata: dict
            Expects the content of `doc_meta.epub_metadata` from digitizer output.
        mets_alto_meta: List[str]
            Expects the content of `doc_meta.mets_alto_metadata` from digitizer output.
        verify_texts_with_llm: bool
            If enabled, each text is passed to text classifier agent first
            and only texts classified as metadata blocks are passed to
            meta extractor(s).
        n_trials: int
            Indicates how many trials to run for predicting metadata with LlamaExtractor
            for the same text. NB! Setting it higher than 1 has purpose only if temperature
            > 0.
        merge_texts: bool
            If enabled, texts are merged into a single text block before passing
            it to LlamaExtractor. Otherwise texts are passed one by one to LlamaExtractor
            and results are merged afterwards.
        min_ratio: float
            Relevant only if n_trials > 1. Indicates the ratio of times a meta value
            has to be predicted during trials. E.g. if min_ratio = 0.7 and a value is predicted
            2 out of 3 trials, it will not be returned as 2/3 = 0.66 < 0.7.
        add_missing_keys: bool
            If enabled, all possible meta keys are added to the output, even if
            the content has not been extracted.
        detect_text_parts: bool
            If enabled, runs text part detection for detecting conclusions, abstracts etc.
        max_length_per_text: int
            If verify_texts_with_llm is set to False, this param is used for dummy metadata detection -
            if a text is longer than the threshold set with this param, it will not be included
            into Llama input.
        n_first_pages: int
            How many first pages to consider for possible Llama input? NB! Not all of them are
            actually added to the input as the pages are passed through prefiltering.
        n_last_pages: int
            How many last pages to consider for possible Llama input? NB! Not all of them are
            actually added to the input as the pages are passed through prefiltering.
        n_strict_include: int
            Number of pages (out of n_first_pages + n_list_pages set) to pass to Llama
            without additional prefiltering.
        simple: bool
            If enabled, the outputs of titles and authors are simplified (some fields
            necessary mostly for constructing final MARC files are removed).
        llama_api_timeout: int
            Llama API query timeout in seconds.
        validate_llama_output: bool
            If enabled, information detected with Llama-Extractor is validated against the original text.
            If the information cannot be found in the original text, it will be excluded from the output.
        allowed_llama_exceptions: tuple
            Due to a funny there where Celery throws SoftTimeoutExceptions that need to be handled
            directly inside the code block that is at the moment running, the lengthy LLAMA request
            with a wide Exception block becomes an issue, so we pass a list of exceptions that we
            do not suppress but want to pass on.
        verify_llama_request: bool | str:
            Whether to use SSL verification during the LLAMA request and if the path to the certfile necessary for it.
        """
        meta_batches = []
        verified_texts = []
        text_parts = []
        meta = {}
        epub_meta = {}
        raw_meta = {}
        extractor_type = []

        success = None
        if epub_metadata:
            extractor_type.append(ExtractorType.EPUB)
            LOGGER.info(f"Using {ExtractorType.EPUB} for metadata extraction.", extra=self.extra_logging)
            meta: dict = self.epub_meta_extractor.extract_meta(
                epub_metadata=epub_metadata,
                simple=simple,
                language=language
            )
            LOGGER.debug(f"Meta detected with {extractor_type}: {meta}", extra=self.extra_logging)

            success = False


        elif mets_alto_metadata:
            mets_alto_meta_extractor = MetsAltoMetaExtractor(
                mets_alto_metadata=mets_alto_metadata,
                language=language,
                texts=texts
            )
            extractor_type.append(ExtractorType.METS_ALTO)
            LOGGER.info(f"Using {ExtractorType.METS_ALTO} for metadata extraction.", extra=self.extra_logging)
            meta: dict = mets_alto_meta_extractor.extract_meta(simple=simple)
            success = True

        if not success:
            extractor_type.append(ExtractorType.LLAMA)
            if success != None:
                epub_meta = deepcopy(meta)

            LOGGER.info(f"Using {ExtractorType.LLAMA} for metadata extraction.", extra=self.extra_logging)

            raw_texts = [
                doc.get("text") for doc in texts
                if doc.get("text").strip()
            ]

            verified_texts = self._construct_llama_input(
                texts=raw_texts,
                use_llm=verify_texts_with_llm,
                max_length_per_text=max_length_per_text,
                n_first_pages=n_first_pages,
                n_last_pages=n_last_pages,
                n_strict_include=n_strict_include,
            )

            if merge_texts:
                text_str = "\n\n".join(verified_texts)
                if text_str:
                    verified_texts = [text_str]
                else:
                    verified_texts = []

                LOGGER.debug(f"Constructed Llama input of size {len(text_str)} characters.", extra=self.extra_logging)

            for text in verified_texts:
                LOGGER.debug(f"Extracting meta from text: {text}", extra=self.extra_logging)
                for trial_nr in range(n_trials):
                    try:
                        LOGGER.debug(
                            f"Trial nr {trial_nr}. Extracting information with Llama agent " \
                            f"from text '{text[:20]}...'.",
                            extra=self.extra_logging
                        )
                        meta_batch = self.meta_agent.extract(
                            text=text,
                            timeout=llama_api_timeout,
                            verify_llama_request=verify_llama_request
                        )
                        LOGGER.debug(f"Raw LLM output: {meta_batch}", extra=self.extra_logging)
                        meta_batches.append(meta_batch)

                    except allowed_llama_exceptions:
                        raise

                    except Exception as e:
                        LOGGER.exception(
                            f"Extracting meta information from text: {text[:50]} " \
                            f"failed with error: {e}.",
                            extra=self.extra_logging
                        )

        # Detect and add text parts
        if detect_text_parts:

            if ExtractorType.EPUB in extractor_type:
                LOGGER.debug(f"Detecting text parts with EPUB text part classifier.", extra=self.extra_logging)
                text_parts = self.epub_text_part_classifier.get_parts_of_text(
                    digitized_texts=texts,
                    epub_meta=meta
                )
                meta[MetaField.TEXT_PARTS] = text_parts

            elif ExtractorType.METS_ALTO in extractor_type:
                LOGGER.debug(f"Detecting text parts with METS/ALTO text part classifier.", extra=self.extra_logging)
                text_parts = self.mets_alto_text_part_classifier.get_parts_of_text(
                    digitized_texts=texts,
                    mets_alto_meta=meta
                )
                meta[MetaField.TEXT_PARTS] = text_parts
            else:
                LOGGER.debug(f"Detecting text parts with Regex text part classifier.", extra=self.extra_logging)
                text_parts = self._get_text_parts(texts)

        if ExtractorType.LLAMA in extractor_type:
            meta = Meta(
                meta_batches=meta_batches,
                text_parts=text_parts,
                min_ratio=min_ratio,
                add_missing_keys=add_missing_keys,
                simple=simple,
                language=language,
                extra_logging=self.extra_logging
            ).to_dict()

            if validate_llama_output:
                meta = self.llama_validator.filter_false_positives(
                    text="\n".join(verified_texts),
                    llama_output=meta
                )

        if ExtractorType.LLAMA in extractor_type:

            if not verified_texts:
                first_pages = texts[:n_first_pages]
                last_pages = texts[-n_last_pages:] if n_last_pages > 0 else []
                relevant_texts = first_pages + last_pages
                verified_texts = [
                    text.get("text")
                    for text in relevant_texts
                ]

            text = "\n".join(verified_texts)
            if ExtractorType.EPUB in extractor_type:
                merger = EpubLlamaMerger(
                    epub_meta=epub_meta,
                    llama_meta=meta,
                    text=text,
                    field_method_map=EPUB_FIELD_METHOD_MAP,
                    default_merge_method=DEFAULT_EPUB_MERGE_METHOD,
                    extra_logging=self.extra_logging
                )
                meta = merger.merge()

            else:
                merger = EpubLlamaMerger(
                    epub_meta=epub_meta,
                    llama_meta=meta,
                    text=text,
                    field_method_map=LLAMA_FIELD_METHOD_MAP,
                    default_merge_method=DEFAULT_LLAMA_MERGE_METHOD,
                    extra_logging=self.extra_logging
                )
                meta = merger.merge()

        # Included texts are expected parts of meta extractor output in
        # text part classifiers, but once text parts are detected (or ignored),
        # they no longer serve purpose in the output
        meta.pop(MetaField.INCLUDED_TEXT, {})

        # Add chapter list, if table of contents is found
        meta[MetaField.TABLE_OF_CONTENTS] = TableOfContents(extracted_meta=meta,
                                                            extra_logging=self.extra_logging).table_of_contents

        # Filter out table of contents in text parts
        filtered_text_parts = []
        for text_part in text_parts:
            if text_part["text_type"] == TextPartLabel.TABLE_OF_CONTENTS:
                continue
            else:
                filtered_text_parts.append(text_part)

        meta[MetaField.TEXT_PARTS] = filtered_text_parts
        meta[MetaField.HOST_ENTRY] = self._get_host_entry(meta)

        series_info = self._get_series_info(meta)

        series_fields = [MetaField.SERIES_ISSN, MetaField.SERIES_NAME, MetaField.SERIES_VOLUME]
        meta[MetaField.SERIES] = series_info
        for field in series_fields:
            meta.pop(field, "")

        if MetaField.ISSUE_TYPE not in meta:
            meta[MetaField.ISSUE_TYPE] = DEFAULT_ISSUE_TYPE

        result = {
            "meta": meta,
            "extractor": extractor_type
        }
        return result



from dataclasses import dataclass

@dataclass(frozen=True)
class TextBlock:
    REGULAR: str = "regular text block"
    METADATA: str = "metadata text block"
    TITLE_PAGE: str = "title page"
    ABSTRACT: str = "abstract text block"
    SUMMARY: str = "summary text block"


@dataclass(frozen=True)
class IssueType:
    NEWSPAPER: str = "Ajaleht"
    JOURNAL: str = "Ajakiri"
    PERIODICAL: str = "Jätkväljaanne"
    BOOK: str = "Raamat"

@dataclass(frozen=True)
class ExtractorType:
    LLAMA: str = "Llama-Extractor"
    EPUB: str = "EPUB-Extractor"
    METS_ALTO: str = "METS/ALTO-Extractor"

@dataclass(frozen=True)
class TextPartLabel:
    TABLE_OF_CONTENTS: str = "Sisukord"
    ABSTRACT: str = "Abstrakt"
    CONCLUSION: str = "Kokkuvõte"
    OTHER: str = "Muu"

@dataclass(frozen=True)
class AuthorField:
    AUTHOR: str = "author"
    UNKNOWN: str = "Teadmata"

@dataclass(frozen=True)
class AuthorType:
    PER: str = "PER"
    ORG: str = "ORG"
    UNK: str = ""


@dataclass(frozen=True)
class TitleType:
    AUTHOR_WITHOUT_TITLE: str = "pealkirjata autor"
    NORMALIZED_TITLE: str = "normitud eelispealkiri"
    TITLE: str = "väljaandes esitatud kujul põhipealkiri"
    PARALLEL_TITLE: str = "rööppealkiri"
    ADDITIONAL_TITLE: str = "alampealkiri"
    METS_TITLE: str = "väljaandes esitatud kujul põhipealkiri"

@dataclass(frozen=True)
class TitleTypeSimple:
    TITLE: str = "main_title"
    ADDITIONAL_TITLE: str = "subtitle"
    METS_TITLE: str = "article_or_chapter_title"
    PARALLEL_TITLE: str = "parallel_title"


@dataclass(frozen=True)
class AuthorNameOrder:
    FIRST_NAME_FIRST: int = 0
    LAST_NAME_FIRST: int = 1
    FAMILY_NAME_FIRST: int = 3

@dataclass(frozen=True)
class DataRestrictions:
    ISBN_MIN_LENGTH: int = 10
    ISBN_MAX_LENGTH: int = 13
    ISSN_LENGTH: int = 8
    YEAR_LENGTH: int = 4

@dataclass(frozen=True)
class MetaYearField:
    DISTRIBUTION_DATE: str = "distribution date"
    PUBLICATION_DATE: str = "publication date"
    MANUFACTURE_DATE: str = "manufacture date"

@dataclass(frozen=True)
class SeriesField:
    ISSN: str = "issn"
    STATEMENT: str = "name"
    VOLUME: str = "volume"

@dataclass(frozen=True)
class MetaField:
    TITLES: str = "titles"
    TITLE: str = "title"
    SUBTITLE: str = "title remainder"
    TITLE_PART_NR: str = "title part nr"
    TITLE_PART_NAME: str = "title part name"
    TITLE_VARFORM: str = "title varform"
    ISBN: str = "isbn"
    ISSN: str = "issn"
    EDITION_INFO: str = "edition info/number"
    DISTRIBUTION_PLACE: str = "distribution_place"
    DISTRIBUTER_NAME: str = "distributer name"
    DISTRIBUTION_DATE: str = "distribution date"
    MANUFACTURE_DATE: str = "manufacture date"
    MANUFACTURER: str = "manufacturer"
    MANUFACTURE_PLACE: str = "manufacture place"
    PUBLICATION_DATE: str = "publication_date"
    PUBLISHER: str = "publisher"
    PUBLICATION_PLACE: str = "publication place"
    COUNTRY_FROM_008: str = "country from 008"
    TEXT_TYPE: str = "text_type"
    AUTHORS: str = "authors"
    TEXT_PARTS: str = "text_parts"
    ISSUE_TYPE: str = "issue_type"
    UDK: str = "udk"
    UDC: str = "udc"
    UNK: str = "unknown"
    INCLUDED_TEXT: str = "description"
    TABLE_OF_CONTENTS: str = "table_of_contents"
    SECTIONS: str = "sections"
    HOST_ENTRY: str = "host_entry"
    SERIES_ISSN: str = "series_issn"
    SERIES_NAME: str = "series_statement"
    SERIES_VOLUME: str = "series_volume"
    SERIES: str = "series"


@dataclass(frozen=True)
class HostEntryField:
    NAME: str = "name"
    PUBLICATION_DATE: str = "publication_date"
    PART_NUMBER: str = "part_number"


@dataclass(frozen=True)
class MergeMethod:
    LLAMA_HARD: str = "LLama-Extractor HARD"
    EPUB_HARD: str = "EPUB-Extractor HARD"
    LLAMA_SOFT: str = "Llama-Extractor SOFT"
    EPUB_SOFT: str = "EPUB-Extractor SOFT"
    COMBO: str = "LLama + EPUB combination"
    REGEX: str = "Regex pattern"
